@@ -1,11 +1,11 @@
-;;; swiper.el --- Isearch with an overview. Oh, man! -*- lexical-binding: t -*-
+;;; swiper.el --- Isearch with an overview.  Oh, man! -*- lexical-binding: t -*-
 
-;; Copyright (C) 2015-2021 Free Software Foundation, Inc.
+;; Copyright (C) 2015-2023 Free Software Foundation, Inc.
 
 ;; Author: Oleh Krehel <ohwoeowho@gmail.com>
 ;; URL: https://github.com/abo-abo/swiper
-;; Version: 0.13.4
-;; Package-Requires: ((emacs "24.5") (ivy "0.13.4"))
+;; Version: 0.14.0
+;; Package-Requires: ((emacs "24.5") (ivy "0.14.0"))
 ;; Keywords: matching
 
 ;; This file is part of GNU Emacs.
@@ -34,6 +34,7 @@
 
 ;;; Code:
 
+(require 'cl-lib)
 (require 'ivy)
 
 (defgroup swiper nil
@@ -332,19 +333,28 @@ If the input is empty, select the previous history element instead."
            (avy-push-mark))
       (avy--done))))
 
+(defun swiper--avy-index (pos)
+  "Return `ivy--index' for `avy' candidate at minibuffer POS."
+  ;; Position in original buffer.
+  (let ((opos (get-text-property pos 'point)))
+    (or
+     ;; Find `swiper-isearch' index based on buffer position.
+     (and opos (cl-position opos ivy--all-candidates))
+     ;; Find `swiper' index based on line number.
+     (let ((nlines (count-lines (point-min) (point-max))))
+       (+ (car (ivy--minibuffer-index-bounds
+                ivy--index ivy--length ivy-height))
+          (line-number-at-pos pos)
+          (if (or (= nlines (1+ ivy-height))
+                  (< ivy--length ivy-height))
+              0
+            (- ivy-height nlines))
+          -2)))))
+
 (defun swiper--avy-goto (candidate)
   (cond ((let ((win (cdr-safe candidate)))
            (and win (window-minibuffer-p win)))
-         (let ((nlines (count-lines (point-min) (point-max))))
-           (ivy-set-index
-            (+ (car (ivy--minibuffer-index-bounds
-                     ivy--index ivy--length ivy-height))
-               (line-number-at-pos (car candidate))
-               (if (or (= nlines (1+ ivy-height))
-                       (< ivy--length ivy-height))
-                   0
-                 (- ivy-height nlines))
-               -2)))
+         (setq ivy--index (swiper--avy-index (car candidate)))
          (ivy--exhibit)
          (ivy-done)
          (ivy-call))
@@ -355,16 +365,19 @@ If the input is empty, select the previous history element instead."
 
 ;;;###autoload
 (defun swiper-avy ()
-  "Jump to one of the current swiper candidates."
+  "Jump to one of the current swiper candidates with `avy'."
   (interactive)
   (unless (require 'avy nil 'noerror)
-    (error "Package avy isn't installed"))
+    (user-error "Package avy isn't installed"))
   (cl-case (length ivy-text)
     (0
      (user-error "Need at least one char of input"))
     (1
-     (let ((swiper-min-highlight 1))
-       (swiper--update-input-ivy))))
+     ;; FIXME: `swiper--update-input-ivy' expects string candidates,
+     ;; but `swiper-isearch' now uses buffer positions.
+     (when (stringp (ivy-state-current ivy-last))
+       (let ((swiper-min-highlight 1))
+         (swiper--update-input-ivy)))))
   (swiper--avy-goto (swiper--avy-candidate)))
 
 (declare-function mc/create-fake-cursor-at-point "ext:multiple-cursors-core")
@@ -561,8 +574,8 @@ numbers; replaces calculating the width from buffer line count."
             (cl-incf line-number))
           (nreverse candidates))))))
 
-(defvar swiper--opoint 1
-  "The point when `swiper' starts.")
+(defvar swiper--opoint nil
+  "Value of point when `swiper' or `swiper-isearch' starts.")
 
 ;;;###autoload
 (defun swiper-backward (&optional initial-input)
@@ -804,7 +817,7 @@ When non-nil, INITIAL-INPUT is the initial search pattern."
                    (if initial-input
                        (cl-position-if
                         (lambda (x)
-                          (= (1+ preselect) (swiper--line-number x)))
+                          (<= (1+ preselect) (swiper--line-number x)))
                         (progn
                           (setq ivy--old-re nil)
                           (ivy--filter initial-input candidates)))
@@ -954,8 +967,9 @@ the face, window and priority of the overlay."
                       (line-move (1- num))
                     (forward-line (1- num))))
                 (if (and (equal ivy-text "")
-                         (>= swiper--opoint (line-beginning-position))
-                         (<= swiper--opoint (line-end-position)))
+                         (<= (line-beginning-position)
+                             swiper--opoint
+                             (line-end-position)))
                     (goto-char swiper--opoint)
                   (if (eq swiper--current-line num)
                       (when swiper--current-match-start
@@ -1114,7 +1128,8 @@ WND, when specified is the window."
                      #'line-move
                    #'forward-line)
                  ln)
-        (when (and (re-search-forward re (line-end-position) t) swiper-goto-start-of-match)
+        (when (and (re-search-forward re (line-end-position) t)
+                   swiper-goto-start-of-match)
           (goto-char (match-beginning 0)))
         (swiper--ensure-visible)
         (swiper--maybe-recenter)
@@ -1372,8 +1387,8 @@ See `ivy-format-functions-alist' for further information."
               (invisible-p (overlay-get ov 'invisible)))
             (overlays-at (point))))))
 
-(defvar swiper--isearch-backward nil)
-(defvar swiper--isearch-start-point nil)
+(defvar swiper--isearch-backward nil
+  "Non-nil when performing `swiper-isearch-backward'.")
 
 (defun swiper--isearch-function-1 (re backward)
   (unless (string= re ".")
@@ -1398,24 +1413,20 @@ See `ivy-format-functions-alist' for further information."
         (nreverse cands)))))
 
 (defun swiper--isearch-next-item (re cands)
-  (if swiper--isearch-backward
-      (or
-       (cl-position-if
-        (lambda (x)
-          (and
-           (< x swiper--isearch-start-point)
-           (eq 0 (string-match-p
-                  re
-                  (buffer-substring-no-properties
-                   x swiper--isearch-start-point)))))
-        cands
-        :from-end t)
-       0)
-    (or
-     (cl-position-if
-      (lambda (x) (> x swiper--isearch-start-point))
-      cands)
-     0)))
+  (or (if swiper--isearch-backward
+          (save-excursion
+            ;; Match RE starting at each position in CANDS.
+            (setq re (concat "\\=\\(?:" re "\\)"))
+            (cl-position-if
+             (lambda (x)
+               (when (< x swiper--opoint)
+                 (goto-char x)
+                 ;; Note: Not quite the same as `looking-at' + `match-end'.
+                 (re-search-forward re swiper--opoint t)))
+             cands
+             :from-end t))
+        (cl-position swiper--opoint cands :test #'<))
+      0))
 
 (defun swiper--isearch-filter-ignore-order (re-full cands)
   (let (filtered-cands)
@@ -1473,11 +1484,30 @@ that we search only for one character."
     (dolist (re (swiper--positive-regexps))
       (swiper--add-overlays re))))
 
+(defun swiper--isearch-candidate-pos (cand)
+  "Return the buffer position of `swiper-isearch' CAND, or nil."
+  (cond ((integer-or-marker-p cand) cand)
+        ((and (stringp cand) (> (length cand) 0))
+         (get-text-property 0 'point cand))))
+
+(defun swiper--isearch-candidate-string (cand)
+  "Return full match of `swiper-isearch' candidate CAND.
+Signal an error on failure."
+  ;; FIXME: Better way of getting current candidate?
+  (or (let ((pos (swiper--isearch-candidate-pos cand))
+            (re (ivy-re-to-str ivy-regex)))
+        (save-match-data
+          (save-excursion
+            (and pos (goto-char pos)
+                 (if (or swiper--isearch-backward swiper-goto-start-of-match)
+                     (looking-at re)
+                   (looking-back re (point-min)))
+                 (match-string 0)))))
+      (error "Could not extract `swiper-isearch' candidate: %S" cand)))
+
 (defun swiper-isearch-action (x)
   "Move to X for `swiper-isearch'."
-  (if (or (numberp x)
-          (and (> (length x) 0)
-               (setq x (get-text-property 0 'point x))))
+  (if (setq x (swiper--isearch-candidate-pos x))
       (with-ivy-window
         (goto-char x)
         (when (and (or (eq this-command 'ivy-previous-line-or-history)
@@ -1487,7 +1517,11 @@ that we search only for one character."
           (goto-char (match-beginning 0)))
         (funcall isearch-filter-predicate (point) (1+ (point)))
         (swiper--maybe-recenter)
-        (if (eq ivy-exit 'done)
+        (if (or (eq ivy-exit 'done)
+                ;; FIXME: With the default action 'M-o o', `ivy-exit' remains
+                ;; nil for some reason, so check `this-command' instead to
+                ;; tell whether we're "done".
+                (eq this-command #'ivy-dispatching-done))
             (progn
               (swiper--push-mark)
               (swiper--remember-search-history (ivy--regex ivy-text)))
@@ -1504,9 +1538,38 @@ that we search only for one character."
     (line-beginning-position) (line-end-position)))
   (goto-char swiper--opoint))
 
-(ivy-add-actions 'swiper-isearch '(("w" swiper-action-copy "copy")))
+(defun swiper-isearch-action-copy (cand)
+  "Save `swiper-isearch' candidate CAND to `kill-ring'.
+Return to original position."
+  (unwind-protect
+      (kill-new (swiper--isearch-candidate-string cand))
+    ;; In case of unexpected error.
+    (goto-char swiper--opoint)))
+
+(defun swiper-isearch-action-insert (cand)
+  "Insert `swiper-isearch' candidate CAND where invoked.
+This cannot currently be called repeatedly without exiting
+completion."
+  (goto-char swiper--opoint)
+  (unwind-protect
+      ;; FIXME: This seems to invalidate many cached buffer positions, thus
+      ;; breaking `ivy-dispatching-call'.
+      (insert (swiper--isearch-candidate-string cand))
+    ;; In case of unexpected error.
+    (goto-char swiper--opoint)))
+
+(ivy-add-actions 'swiper-isearch '(("w" swiper-isearch-action-copy "copy")))
+(ivy-add-actions 'swiper-isearch '(("i" swiper-isearch-action-insert "insert")))
 (ivy-add-actions 'swiper '(("w" swiper-action-copy "copy")))
 
+(defun swiper--isearch-insert-current ()
+  "Replace minibuffer contents with the current candidate.
+Like `ivy-insert-current', but tailored for `swiper-isearch'."
+  (interactive)
+  (delete-minibuffer-contents)
+  (let ((cur (ivy-state-current ivy-last)))
+    (insert (with-ivy-window (swiper--isearch-candidate-string cur)))))
+
 (defun swiper-isearch-thing-at-point ()
   "Insert `symbol-at-point' into the minibuffer of `swiper-isearch'.
 When not running `swiper-isearch' already, start it."
@@ -1536,10 +1599,20 @@ When not running `swiper-isearch' already, start it."
           (setq thing (ivy-thing-at-point))))
       (swiper-isearch thing))))
 
+(defun swiper-isearch-C-r (&optional arg)
+  "Move cursor vertically up ARG candidates.
+When the input is empty, browse the search history instead."
+  (interactive "p")
+  (if (string= ivy-text "")
+      (ivy-reverse-i-search)
+    (ivy-previous-line arg)))
+
 (defvar swiper-isearch-map
   (let ((map (make-sparse-keymap)))
     (set-keymap-parent map swiper-map)
-    (define-key map (kbd "M-n") 'swiper-isearch-thing-at-point)
+    (define-key map [remap ivy-insert-current] #'swiper--isearch-insert-current)
+    (define-key map (kbd "M-n") #'swiper-isearch-thing-at-point)
+    (define-key map (kbd "C-r") #'swiper-isearch-C-r)
     map)
   "Keymap for `swiper-isearch'.")
 
@@ -1648,7 +1721,6 @@ When not running `swiper-isearch' already, start it."
 (defun swiper--isearch-init ()
   "Initialize `swiper-isearch'."
   (swiper--init)
-  (setq swiper--isearch-start-point (point))
   (swiper-font-lock-ensure))
 
 (defun swiper--isearch-unwind ()
